extern "C" {
#include <stdio.h>
#include <assert.h>
#include <stdbool.h>
#include "math.h" // sqrt, ieee, NAN, isnan
}

/****************************************************************

 A complete fast single-pass implementation of a bivariate OLS
 regression written in Mongrel C/C++.

 The underlying functionality and variable names are explained in
 most statistics textbooks.  a is the intercept, b is the slope;

 test and compile as 'g++ -Wall -o biols biols.cc' where biols.cc contains
 
     #include <biols.hh>
     int main() { BIOLS bcalc; bcalc.test(); }

  (C) Ivo Welch, Nov 2017.  Released GPL Copyleft.
 *****************************************************************/

class BIOLS {
protected:
  double sumxx, sumx, sumxy, sumy, sumyy; int N; int NALL;
  //  If sum* become very large, this can display numerical instability.

  static bool issimilar(const double x, const double y) { return fabs( x - y) < 0.00001; }

public:
  BIOLS() { init(); }

  void init() { sumxx=sumx=sumxy=sumy=sumyy=0.0; N=NALL=0; }

  void add( const double y, const double x ) {
    ++NALL;
    if (isnan( y )) return; if (isnan( x )) return;
    sumxx+= x*x; sumx+=x; sumy+= y; sumxy+= x*y; sumyy+= y*y; ++N;
  }

  int n() const { return N; }
  double rho() const { return (N<=1) ? NAN : (-1)*((sumy*sumx-sumxy*N)/sqrt((sumx*sumx-sumxx*N)*(sumy*sumy-sumyy*N))); }
  double cor() const { return rho(); }
  double xavg() const { return sumx/N; }
  double yavg() const { return sumy/N; }
  double xvar() const { return (sumxx-sumx*sumx/N)/(N-1); } double xsdv() const { return sqrt(xvar()); }
  double yvar() const { return (sumyy-sumy*sumy/N)/(N-1); } double ysdv() const { return sqrt(yvar()); }
  double xvarp() const { return (sumxx-sumx*sumx/N)/N; } double xsdvp() const { return sqrt(xvarp()); }
  double yvarp() const { return (sumyy-sumy*sumy/N)/N; } double ysdvp() const { return sqrt(yvarp()); }

  double rsq() const { double r= rho(); return r*r; }
  double adjrsq() const { return (N<=2) ? NAN : 1.0-(double)(N-1.0)/(N-2.0)*(1-rsq()); }
  double s() const { return (N<=2) ? NAN : sqrt( evar()*(N-1.0)/(N-2.0) ); }
  double evar() const { return (N<=2) ? NAN : (1.0-rsq())*yvar(); } double esdv() const { return sqrt(evar()); }

  double b() const { return (N<=1) ? NAN : (sumy*sumx-sumxy*N)/(sumx*sumx-sumxx*N); }
  double a() const { return (N<=1) ? NAN : yavg()-b()*xavg(); }

  double asdv() const { return (N<=2) ? NAN : s()*sqrt( 1.0/N*(1.0+xavg()*xavg()/xvarp()) ); }
  double bsdv() const { return (N<=2) ? NAN : s()/sqrt(N*xvarp()); }

  double aT() const { return a()/asdv(); }
  double bT() const { return b()/bsdv(); }
  double F() const { return bT()*bT(); }

public:
  // sugar; could be inherited instead

  void print(const char *optnames[2] =NULL, FILE *f =stdout) const {
    const char *defaultnames[2]= { "y", "x" };
    const char **names;
    names= (optnames) ? optnames : defaultnames;
    fprintf(f, "----------------\n");
    fprintf(f, "%s = %8.3f + %8.3f * %s\n", names[0], a(), b(), names[1]);
    fprintf(f, "se (%8.3f) (%8.3f)\n", asdv(), bsdv());
    fprintf(f, "T (%8.3f) (%8.3f)\n", aT(), bT());
    fprintf(f, "\ts=%.3f; cor=%.3f; R^2=%.4f (%.5f).\n", s(), cor(), rsq(), adjrsq());
    fprintf(f, "n=%d(%d).  means: %s=%g(sd=%g) %s=%g(sd=%g)\n", N, NALL, names[0], xavg(), xsdv(), names[1], yavg(), ysdv() );
    fprintf(f, "----------------\n");
  }

  void test() {
    for (int i=1; i<=20; ++i) { add( (double)i*i, (double)i ); }
    print(NULL); // no specific names by default

    assert( issimilar(xavg(), 10.5) );
    assert( issimilar(yavg(), 143.5) );
    assert( issimilar(xsdv(), 5.916079783) );
    assert( issimilar(ysdv(), 127.9023065) );
    assert( issimilar(b(), 21.0) );
    assert( issimilar(a(), -77.0) );
    assert( issimilar(cor(), 0.9713482022) );
    assert( issimilar(rsq(), 0.9435173) );
    assert( issimilar(adjrsq(), 0.9403794 ) );
    assert( issimilar( esdv(), 30.39736831 ) );
    assert( issimilar( s(), 31.23033 ) );
    assert( issimilar(asdv(), 14.507469341 ) );
    assert( issimilar(bsdv(), 1.211060142 ) );
  }

};

